"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.handler = exports.getExecutionDuration = exports.createEmptyRecursiveSummary = exports.writeRecursiveSummary = exports.getResumedPackageChunks = exports.help = exports.cliOptionsTypes = exports.rcOptionsTypes = exports.commandNames = exports.shorthands = void 0;
const path_1 = __importDefault(require("path"));
const cli_utils_1 = require("@pnpm/cli-utils");
const config_1 = require("@pnpm/config");
const lifecycle_1 = require("@pnpm/lifecycle");
const logger_1 = require("@pnpm/logger");
const read_project_manifest_1 = require("@pnpm/read-project-manifest");
const sort_packages_1 = require("@pnpm/sort-packages");
const execa_1 = __importDefault(require("execa"));
const p_limit_1 = __importDefault(require("p-limit"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const render_help_1 = __importDefault(require("render-help"));
const existsInDir_1 = require("./existsInDir");
const makeEnv_1 = require("./makeEnv");
const run_1 = require("./run");
const error_1 = require("@pnpm/error");
const write_json_file_1 = __importDefault(require("write-json-file"));
exports.shorthands = {
    parallel: run_1.shorthands.parallel,
    c: '--shell-mode',
};
exports.commandNames = ['exec'];
function rcOptionsTypes() {
    return {
        ...(0, pick_1.default)([
            'bail',
            'sort',
            'use-node-version',
            'unsafe-perm',
            'workspace-concurrency',
        ], config_1.types),
        'shell-mode': Boolean,
        'resume-from': String,
        'report-summary': Boolean,
    };
}
exports.rcOptionsTypes = rcOptionsTypes;
const cliOptionsTypes = () => ({
    ...rcOptionsTypes(),
    recursive: Boolean,
    reverse: Boolean,
});
exports.cliOptionsTypes = cliOptionsTypes;
function help() {
    return (0, render_help_1.default)({
        description: 'Run a shell command in the context of a project.',
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    run_1.PARALLEL_OPTION_HELP,
                    {
                        description: 'Run the shell command in every package found in subdirectories \
or every workspace package, when executed inside a workspace. \
For options that may be used with `-r`, see "pnpm help recursive"',
                        name: '--recursive',
                        shortAlias: '-r',
                    },
                    {
                        description: 'If exist, runs file inside of a shell. \
Uses /bin/sh on UNIX and \\cmd.exe on Windows. \
The shell should understand the -c switch on UNIX or /d /s /c on Windows.',
                        name: '--shell-mode',
                        shortAlias: '-c',
                    },
                    run_1.RESUME_FROM_OPTION_HELP,
                    run_1.REPORT_SUMMARY_OPTION_HELP,
                ],
            },
        ],
        url: (0, cli_utils_1.docsUrl)('exec'),
        usages: ['pnpm [-r] [-c] exec <command> [args...]'],
    });
}
exports.help = help;
function getResumedPackageChunks({ resumeFrom, chunks, selectedProjectsGraph, }) {
    const resumeFromPackagePrefix = Object.keys(selectedProjectsGraph)
        .find((prefix) => selectedProjectsGraph[prefix]?.package.manifest.name === resumeFrom);
    if (!resumeFromPackagePrefix) {
        throw new error_1.PnpmError('RESUME_FROM_NOT_FOUND', `Cannot find package ${resumeFrom}. Could not determine where to resume from.`);
    }
    const chunkPosition = chunks.findIndex(chunk => chunk.includes(resumeFromPackagePrefix));
    return chunks.slice(chunkPosition);
}
exports.getResumedPackageChunks = getResumedPackageChunks;
async function writeRecursiveSummary(opts) {
    await (0, write_json_file_1.default)(path_1.default.join(opts.dir, 'pnpm-exec-summary.json'), {
        executionStatus: opts.summary,
    });
}
exports.writeRecursiveSummary = writeRecursiveSummary;
function createEmptyRecursiveSummary(chunks) {
    return chunks.flat().reduce((acc, prefix) => {
        acc[prefix] = { status: 'queued' };
        return acc;
    }, {});
}
exports.createEmptyRecursiveSummary = createEmptyRecursiveSummary;
function getExecutionDuration(start) {
    const end = process.hrtime(start);
    return (end[0] * 1e9 + end[1]) / 1e6;
}
exports.getExecutionDuration = getExecutionDuration;
async function handler(opts, params) {
    // For backward compatibility
    if (params[0] === '--') {
        params.shift();
    }
    const limitRun = (0, p_limit_1.default)(opts.workspaceConcurrency ?? 4);
    let chunks;
    if (opts.recursive) {
        chunks = opts.sort
            ? (0, sort_packages_1.sortPackages)(opts.selectedProjectsGraph)
            : [Object.keys(opts.selectedProjectsGraph).sort()];
        if (opts.reverse) {
            chunks = chunks.reverse();
        }
    }
    else {
        chunks = [[opts.dir]];
        const project = await (0, read_project_manifest_1.tryReadProjectManifest)(opts.dir);
        if (project.manifest != null) {
            opts.selectedProjectsGraph = {
                [opts.dir]: {
                    dependencies: [],
                    package: {
                        ...project,
                        dir: opts.dir,
                    },
                },
            };
        }
    }
    if (opts.resumeFrom) {
        chunks = getResumedPackageChunks({
            resumeFrom: opts.resumeFrom,
            chunks,
            selectedProjectsGraph: opts.selectedProjectsGraph,
        });
    }
    const result = createEmptyRecursiveSummary(chunks);
    const existsPnp = existsInDir_1.existsInDir.bind(null, '.pnp.cjs');
    const workspacePnpPath = opts.workspaceDir && await existsPnp(opts.workspaceDir);
    let exitCode = 0;
    for (const chunk of chunks) {
        await Promise.all(chunk.map(async (prefix) => limitRun(async () => {
            result[prefix].status = 'running';
            const startTime = process.hrtime();
            try {
                const pnpPath = workspacePnpPath ?? await existsPnp(prefix);
                const extraEnv = {
                    ...opts.extraEnv,
                    ...(pnpPath ? (0, lifecycle_1.makeNodeRequireOption)(pnpPath) : {}),
                };
                const env = (0, makeEnv_1.makeEnv)({
                    extraEnv: {
                        ...extraEnv,
                        PNPM_PACKAGE_NAME: opts.selectedProjectsGraph[prefix]?.package.manifest.name,
                    },
                    prependPaths: [
                        './node_modules/.bin',
                        ...opts.extraBinPaths,
                    ],
                    userAgent: opts.userAgent,
                });
                await (0, execa_1.default)(params[0], params.slice(1), {
                    cwd: prefix,
                    env,
                    stdio: 'inherit',
                    shell: opts.shellMode ?? false,
                });
                result[prefix].status = 'passed';
                result[prefix].duration = getExecutionDuration(startTime);
            }
            catch (err) { // eslint-disable-line
                if (!opts.recursive && typeof err.exitCode === 'number') {
                    exitCode = err.exitCode;
                    return;
                }
                logger_1.logger.info(err);
                result[prefix] = {
                    status: 'failure',
                    duration: getExecutionDuration(startTime),
                    error: err,
                    message: err.message,
                    prefix,
                };
                if (!opts.bail) {
                    return;
                }
                if (!err['code']?.startsWith('ERR_PNPM_')) {
                    err['code'] = 'ERR_PNPM_RECURSIVE_EXEC_FIRST_FAIL';
                }
                err['prefix'] = prefix;
                opts.reportSummary && await writeRecursiveSummary({
                    dir: opts.lockfileDir ?? opts.dir,
                    summary: result,
                });
                /* eslint-enable @typescript-eslint/dot-notation */
                throw err;
            }
        })));
    }
    opts.reportSummary && await writeRecursiveSummary({
        dir: opts.lockfileDir ?? opts.dir,
        summary: result,
    });
    (0, cli_utils_1.throwOnCommandFail)('pnpm recursive exec', result);
    return { exitCode };
}
exports.handler = handler;
//# sourceMappingURL=exec.js.map