"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getSpecifiedScripts = exports.runRecursive = void 0;
const path_1 = __importDefault(require("path"));
const cli_utils_1 = require("@pnpm/cli-utils");
const error_1 = require("@pnpm/error");
const lifecycle_1 = require("@pnpm/lifecycle");
const logger_1 = require("@pnpm/logger");
const sort_packages_1 = require("@pnpm/sort-packages");
const p_limit_1 = __importDefault(require("p-limit"));
const realpath_missing_1 = __importDefault(require("realpath-missing"));
const existsInDir_1 = require("./existsInDir");
const exec_1 = require("./exec");
const run_1 = require("./run");
const regexpCommand_1 = require("./regexpCommand");
async function runRecursive(params, opts) {
    const [scriptName, ...passedThruArgs] = params;
    if (!scriptName) {
        throw new error_1.PnpmError('SCRIPT_NAME_IS_REQUIRED', 'You must specify the script you want to run');
    }
    let hasCommand = 0;
    const sortedPackageChunks = opts.sort
        ? (0, sort_packages_1.sortPackages)(opts.selectedProjectsGraph)
        : [Object.keys(opts.selectedProjectsGraph).sort()];
    let packageChunks = opts.reverse ? sortedPackageChunks.reverse() : sortedPackageChunks;
    if (opts.resumeFrom) {
        packageChunks = (0, exec_1.getResumedPackageChunks)({
            resumeFrom: opts.resumeFrom,
            chunks: packageChunks,
            selectedProjectsGraph: opts.selectedProjectsGraph,
        });
    }
    const limitRun = (0, p_limit_1.default)(opts.workspaceConcurrency ?? 4);
    const stdio = !opts.stream &&
        (opts.workspaceConcurrency === 1 ||
            (packageChunks.length === 1 && packageChunks[0].length === 1))
        ? 'inherit'
        : 'pipe';
    const existsPnp = existsInDir_1.existsInDir.bind(null, '.pnp.cjs');
    const workspacePnpPath = opts.workspaceDir && await existsPnp(opts.workspaceDir);
    const requiredScripts = opts.rootProjectManifest?.pnpm?.requiredScripts ?? [];
    if (requiredScripts.includes(scriptName)) {
        const missingScriptPackages = packageChunks
            .flat()
            .map((prefix) => opts.selectedProjectsGraph[prefix])
            .filter((pkg) => getSpecifiedScripts(pkg.package.manifest.scripts ?? {}, scriptName).length < 1)
            .map((pkg) => pkg.package.manifest.name ?? pkg.package.dir);
        if (missingScriptPackages.length) {
            throw new error_1.PnpmError('RECURSIVE_RUN_NO_SCRIPT', `Missing script "${scriptName}" in packages: ${missingScriptPackages.join(', ')}`);
        }
    }
    const result = (0, exec_1.createEmptyRecursiveSummary)(packageChunks);
    for (const chunk of packageChunks) {
        const selectedScripts = chunk.map(prefix => {
            const pkg = opts.selectedProjectsGraph[prefix];
            const specifiedScripts = getSpecifiedScripts(pkg.package.manifest.scripts ?? {}, scriptName);
            if (!specifiedScripts.length) {
                result[prefix].status = 'skipped';
            }
            return specifiedScripts.map(script => ({ prefix, scriptName: script }));
        }).flat();
        await Promise.all(selectedScripts.map(async ({ prefix, scriptName }) => limitRun(async () => {
            const pkg = opts.selectedProjectsGraph[prefix];
            if (!pkg.package.manifest.scripts?.[scriptName] ||
                process.env.npm_lifecycle_event === scriptName &&
                    process.env.PNPM_SCRIPT_SRC_DIR === prefix) {
                return;
            }
            result[prefix].status = 'running';
            const startTime = process.hrtime();
            hasCommand++;
            try {
                const lifecycleOpts = {
                    depPath: prefix,
                    extraBinPaths: opts.extraBinPaths,
                    extraEnv: opts.extraEnv,
                    pkgRoot: prefix,
                    rawConfig: opts.rawConfig,
                    rootModulesDir: await (0, realpath_missing_1.default)(path_1.default.join(prefix, 'node_modules')),
                    scriptsPrependNodePath: opts.scriptsPrependNodePath,
                    scriptShell: opts.scriptShell,
                    shellEmulator: opts.shellEmulator,
                    stdio,
                    unsafePerm: true, // when running scripts explicitly, assume that they're trusted.
                };
                const pnpPath = workspacePnpPath ?? await existsPnp(prefix);
                if (pnpPath) {
                    lifecycleOpts.extraEnv = {
                        ...lifecycleOpts.extraEnv,
                        ...(0, lifecycle_1.makeNodeRequireOption)(pnpPath),
                    };
                }
                const _runScript = run_1.runScript.bind(null, { manifest: pkg.package.manifest, lifecycleOpts, runScriptOptions: { enablePrePostScripts: opts.enablePrePostScripts ?? false }, passedThruArgs });
                await _runScript(scriptName);
                result[prefix].status = 'passed';
                result[prefix].duration = (0, exec_1.getExecutionDuration)(startTime);
            }
            catch (err) { // eslint-disable-line
                logger_1.logger.info(err);
                result[prefix] = {
                    status: 'failure',
                    duration: (0, exec_1.getExecutionDuration)(startTime),
                    error: err,
                    message: err.message,
                    prefix,
                };
                if (!opts.bail) {
                    return;
                }
                err['code'] = 'ERR_PNPM_RECURSIVE_RUN_FIRST_FAIL';
                err['prefix'] = prefix;
                opts.reportSummary && await (0, exec_1.writeRecursiveSummary)({
                    dir: opts.workspaceDir ?? opts.dir,
                    summary: result,
                });
                /* eslint-enable @typescript-eslint/dot-notation */
                throw err;
            }
        })));
    }
    if (scriptName !== 'test' && !hasCommand && !opts.ifPresent) {
        const allPackagesAreSelected = Object.keys(opts.selectedProjectsGraph).length === opts.allProjects.length;
        if (allPackagesAreSelected) {
            throw new error_1.PnpmError('RECURSIVE_RUN_NO_SCRIPT', `None of the packages has a "${scriptName}" script`);
        }
        else {
            logger_1.logger.info({
                message: `None of the selected packages has a "${scriptName}" script`,
                prefix: opts.workspaceDir,
            });
        }
    }
    opts.reportSummary && await (0, exec_1.writeRecursiveSummary)({
        dir: opts.workspaceDir ?? opts.dir,
        summary: result,
    });
    (0, cli_utils_1.throwOnCommandFail)('pnpm recursive run', result);
}
exports.runRecursive = runRecursive;
function getSpecifiedScripts(scripts, scriptName) {
    // if scripts in package.json has script which is equal to scriptName a user passes, return it.
    if (scripts[scriptName]) {
        return [scriptName];
    }
    const scriptSelector = (0, regexpCommand_1.tryBuildRegExpFromCommand)(scriptName);
    // if scriptName which a user passes is RegExp (like /build:.*/), multiple scripts to execute will be selected with RegExp
    if (scriptSelector) {
        const scriptKeys = Object.keys(scripts);
        return scriptKeys.filter(script => script.match(scriptSelector));
    }
    return [];
}
exports.getSpecifiedScripts = getSpecifiedScripts;
//# sourceMappingURL=runRecursive.js.map